from scikit-learn.cluster import KMeans
from scikit-learn.preprocessing import StandardScaler
from scikit-learn.metrics import silhouette_score

import pandas as pd

def cluster_text_positions(text_positions_df):
    clustered_data = []
    
    pages = text_positions_df['page'].unique()
    
    for page in pages:
        page_df = text_positions_df[text_positions_df['page'] == page]
        positions = page_df[['x0', 'y0', 'x1', 'y1']].values
        
        # Standardize the positions
        scaler = StandardScaler()
        positions_scaled = scaler.fit_transform(positions)
        
        # Determine the optimal number of clusters using the Silhouette Score
        silhouette_scores = []
        cluster_range = range(2, min(10, len(page_df)))  # Avoid too many clusters
        
        for n_clusters in cluster_range:
            kmeans = KMeans(n_clusters=n_clusters, random_state=0)
            labels = kmeans.fit_predict(positions_scaled)
            score = silhouette_score(positions_scaled, labels)
            silhouette_scores.append((n_clusters, score))
        
        # Select the number of clusters with the highest silhouette score
        optimal_clusters = max(silhouette_scores, key=lambda x: x[1])[0]
        
        # Perform clustering with the optimal number of clusters
        kmeans = KMeans(n_clusters=optimal_clusters, random_state=0)
        labels = kmeans.fit_predict(positions_scaled)
        
        page_df['cluster'] = labels
        clustered_data.append(page_df)
    
    return pd.concat(clustered_data)


import pandas as pd

from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.pdfpage import PDFPage
from pdfminer.converter import PDFPageAggregator
from pdfminer.layout import LAParams, LTTextBox, LTTextLine, LTChar
from pdfminer.pdfdocument import PDFDocument
from pdfminer.pdfparser import PDFParser
from pdfminer.pdfpage import PDFTextExtractionNotAllowed


def extract_text_with_positions(pdf_path):
    
    rsrcmgr = PDFResourceManager()
    laparams = LAParams()
    device = PDFPageAggregator(rsrcmgr, laparams=laparams)
    interpreter = PDFPageInterpreter(rsrcmgr, device)
    
    with open(pdf_path, 'rb') as f:
        parser = PDFParser(f)
        document = PDFDocument(parser)
        if not document.is_extractable:
            raise PDFTextExtractionNotAllowed
        text_positions = []
        
        for page_number, page in enumerate(PDFPage.create_pages(document)):
            interpreter.process_page(page)
            layout = device.get_result()
            for element in layout:
                if isinstance(element, LTTextBox):
                    for text_line in element:
                        text = ''
                        fontname = ''
                        size = 0
                        for char in text_line:
                            if isinstance(char, LTChar):
                                text += char.get_text()
                                fontname = char.fontname
                                size = char.size
                        text_positions.append({
                            'page': page_number + 1,
                            'text': text.encode('utf-8').decode('utf-8'),
                            'x0': text_line.x0,
                            'y0': text_line.y0,
                            'x1': text_line.x1,
                            'y1': text_line.y1,
                            'height': text_line.height,
                            'width': text_line.width,
                            'fontname': fontname,
                            'size': size
                        })
    return pd.DataFrame(text_positions)

def filter_text_positions(df, header_height, footer_height):
    # Filter out header and footer areas
    df_filtered = df[(df['y0'] > footer_height) & (df['y1'] < (df['y1'].max() - header_height))]
    return df_filtered


import re

def extract_and_order_text(clustered_text_positions_df):
    ordered_texts = {}
    
    pages = clustered_text_positions_df['page'].unique()
    
    for page in pages:
        page_df = clustered_text_positions_df[clustered_text_positions_df['page'] == page]
        clusters = page_df['cluster'].unique()
        
        page_text = []
        
        for cluster in clusters:
            cluster_df = page_df[page_df['cluster'] == cluster]
            
            # Sort text first from left to right, then from top to bottom
            cluster_df = cluster_df.sort_values(by=['y0', 'x0'], ascending=[False, True])
            
            cluster_text = ''.join(cluster_df['text'].tolist())
            page_text.append(cluster_text)
        
        ordered_texts[page] = ' '.join(page_text)
    
    return ordered_texts


def clean_text(text):
    # Remove unwanted characters using regular expressions
    text = re.sub(r'[\t\r\n\xa0\xad]', '', text)
    text = re.sub(r' +', ' ', text)  # Replace multiple spaces with a single space
    text = text.strip()  # Remove leading and trailing spaces
    return text

def clean_ordered_texts(ordered_texts):
    cleaned_texts = {}
    for page, text in ordered_texts.items():
        cleaned_texts[page] = clean_text(text)
    return cleaned_texts




def create_markdown_document(clustered_text_positions_df):
    markdown_document = ""
    
    pages = clustered_text_positions_df['page'].unique()
    
    for page in pages:
        page_df = clustered_text_positions_df[clustered_text_positions_df['page'] == page]
        clusters = page_df['cluster'].unique()
        
        markdown_document += f"## Page {page}\n\n"
        
        for cluster in clusters:
            cluster_df = page_df[page_df['cluster'] == cluster]
            
            # Sort text first from left to right, then from top to bottom
            cluster_df = cluster_df.sort_values(by=['y0', 'x0'], ascending=[False, True])
            
            font_size = None
            cluster_text = ""
            
            for index, row in cluster_df.iterrows():
                # Remove '\r' and '\t'
                text = row['text'].replace("\r", "").replace("\t", "").replace("\xa0", "")
                text = re.sub(' +', ' ', text)
                
                # Round font size to three decimal places
                font_size_rounded = round(row['size'], 3)
                
                # Concatenate subsequent rows with the same font size
                if font_size == font_size_rounded:
                    cluster_text += f"\n{text}"
                else:
                    if font_size:
                        cluster_text += "</span>\n"
                    cluster_text += f"<span>" + text
                    font_size = font_size_rounded
            
            # Close the style tag if there's any text in the cluster
            if font_size:
                cluster_text += "</span>\n"

            
            markdown_document += cluster_text + "\n"
    
    return markdown_document


import numpy as np
import pandas as pd

from .to_txt import extract_text_with_positions,filter_text_positions
from .to_md import create_markdown_document
from .clusterer import cluster_text_positions
from .editor import extract_and_order_text, clean_text, clean_ordered_texts

def pdf_to_str(pdf_path, header_height, footer_height):

    text_positions_df = extract_text_with_positions(pdf_path)
    filtered_text_positions_df = filter_text_positions(text_positions_df, header_height, footer_height)
    clustered_text_positions_df = cluster_text_positions(filtered_text_positions_df)
    ordered_texts = extract_and_order_text(clustered_text_positions_df)
    cleaned_ordered_texts = clean_ordered_texts(ordered_texts)

    return cleaned_ordered_texts

def pdf_to_md(pdf_path, header_height, footer_height):

    text_positions_df = extract_text_with_positions(pdf_path)
    filtered_text_positions_df = filter_text_positions(text_positions_df, header_height, footer_height)
    clustered_text_positions_df = cluster_text_positions(filtered_text_positions_df)
    markdown_export = create_markdown_document(clustered_text_positions_df) 

    return markdown_export

def export_markdown_file(markdown_content, output_path='output.md'):
    with open(output_path, 'w', encoding='utf-8') as file:
        file.write(markdown_content)